import dash
import dash_cytoscape as cyto
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output, State
from typing import Dict, List, Tuple, Union

from .knowledge_graph import search_datasets_for_search_term, RenkuDataSet, RenkuProject

RenkuObject = Union[RenkuDataSet, RenkuProject]

external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']

app = dash.Dash(__name__, title='Renku Knowledge Graph', external_stylesheets=external_stylesheets,
                meta_tags=[{"name": "viewport", "content": "width=device-width"}])


def get_app_layout(graph_elements: List[dict], node_metadata: dict) -> html.Div:
    """
    Builds the dashboard layout, populating it with initial graph data.
    Args:
        graph_elements: Elements to display in the dashboard at start-up. Data structure must match cytoscape.
        node_metadata: Metadata about the nodes in the graph. This data is not displayed, but instead stored in the
        dashboard to be retrieved as the user clicks on elements in the graph.

    Returns: Dashboard layout

    """
    return html.Div(
        [
            html.Div([
                    html.Div([
                            html.H1('🔭 Renku Knowledge Graph Explorer'),
                            dcc.Markdown('Renku is a knowledge infrastructure that accompanies a data scientist '
                                         'throughout all stages of the research lifecycle. '
                                         'At its core, the Renku Knowledge graph links versioned digital objects'
                                         ' within and across collaborative projects. With this explorer, discover'
                                         ' how Renku projects around the world are connected via the shared use of the'
                                         ' same datasets!'),
                            dcc.Markdown('🔍 *Search ideas: COVID, population, kaggle, ...*'),
                            dcc.Input(
                                id='search-box',
                                placeholder='Search datasets...',
                                type='text',
                                value=''
                            ),
                        ],
                        className="pretty_container twelve columns",
                    ),

                ],
                className="row flex-display",
            ),
            html.Div(
                [
                    html.Div(
                        [
                            html.Div([
                                    html.Label('Node Legend:', style={'margin-right': '10px'}),
                                    html.Label('Search term', style={'color': 'red', 'margin-right': '10px'}),
                                    html.Label('Datasets', style={'color': 'blue', 'margin-right': '10px'}),
                                    html.Label('Projects', style={'color': 'orange', 'margin-right': '10px'}),
                                ],
                                className="row flex-display",
                            ),
                            cyto.Cytoscape(
                                id='cytoscape-data-viz',
                                layout={
                                    'name': 'cose',
                                },
                                style={'width': '100%', 'height': '600px'},
                                elements=graph_elements,
                                stylesheet=[
                                    # Group selectors
                                    {
                                        'selector': 'node',
                                        'style': {
                                            'content': 'data(label)',
                                            "font-size": "9px",
                                        }
                                    },

                                    # Class selectors
                                    {
                                        'selector': '.search_term',
                                        'style': {
                                            'background-color': 'red',
                                            'line-color': 'red'
                                        }
                                    },
                                    {
                                        'selector': '.dataset',
                                        'style': {
                                            'background-color': 'blue',
                                            'line-color': 'blue'
                                        }
                                    },
                                    {
                                        'selector': '.project',
                                        'style': {
                                            'background-color': 'orange',
                                            'line-color': 'orange'
                                        }
                                    }
                                ]
                            ),
                            dcc.Markdown("✨ Click on nodes to learn more about them and see what other datasets and"
                                         " projects they're connected to. You can zoom, pan, and drag nodes, too!")
                        ],
                        className="pretty_container eight columns",
                        # style={"border": "2px black solid"}
                    ),
                    html.Div(
                        [
                            dcc.Loading(
                                id="loading-text",
                                children=[
                                    dcc.Markdown(id='text-display'),
                                ]
                            ),
                        ],
                        className="pretty_container four columns"
                    )
                ],
                className="row flex-display"
            ),
            dcc.Store(id='graph_elements', data=graph_elements),
            dcc.Store(id='node_metadata', data=node_metadata),
        ]
    )


@app.callback(
    Output('cytoscape-data-viz', 'elements'),
    [Input('graph_elements', 'data'),
     ])
def update_graph(updated_graph_elements):
    """Whenever the graph_elements store in the dashboard is updated, also update the cytoscape visualization."""
    return updated_graph_elements


@app.callback(
    [Output('graph_elements', 'data'),
     Output('node_metadata', 'data'),
     Output('text-display', 'children')],
    [Input('search-box', 'value'),
     Input('cytoscape-data-viz', 'tapNodeData'),
     State('graph_elements', 'data'),
     State('node_metadata', 'data'),
     ])
def triage_graph_callback(search_text: str, clicked_node_data: dict, graph_elements: List[dict],
                          serialized_node_metadata: List[Dict]) -> Tuple[List[dict], List[dict], str]:
    """
    When the user makes a new search or clicks on a node in the graph, update the graph with new nodes.
    Dash only allows a dashboard element to have 1 callback that updates it. To maintain that, this callback combines
     the search callback and the node-click callback, both of which update the graph.
     See https://dash.plotly.com/advanced-callbacks for more info.

    Args:
        search_text: The text the user typed into the Search box.
        clicked_node_data: Data dictionary for the graph element that the user clicked on.
        graph_elements: The graph elements currently depicted in the graph.
        serialized_node_metadata: Metadata for all nodes currently depicted in the graph, serialized.

    Returns:
        * Updated graph elements. If this callback was triggered by a search, the graph elements are totally replaced
         with new ones. If this callback was triggered by a node-click, additional nodes are added to the current set.
        * node metadata dictionary. Same logic applied as for graph elements. Metadata is serialized for storage in the
         dashboard Store component.
        * metadata info encoded as markdown text, describing the node that the user clicked on (if applicable).

    """
    # determine which input(s) were triggered - search box or node click
    ctx = dash.callback_context
    triggers = [t['prop_id'] for t in ctx.triggered]

    # set return variables to starting values in case nothing is updated
    new_graph_elements = graph_elements
    node_metadata = deserialize_node_metadata(serialized_node_metadata)
    metadata_str = ''

    # If the search box was modified, update the graph with a new search result
    if 'search-box.value' in triggers:
        new_graph_elements, new_node_metadata = update_graph_with_search(search_text)

    # If a node was clicked, (1) show the node's metadata in the right panel, and (2) add the node's related nodes to
    # the graph
    elif 'cytoscape-data-viz.tapNodeData' in triggers:
        clicked_node_id = clicked_node_data['id']
        node_object = node_metadata[clicked_node_id]
        metadata_str = node_object.generate_markdown_description()
        new_graph_elements, new_node_metadata = add_related_nodes_to_graph(clicked_node_id, graph_elements,
                                                                           node_metadata)
    else:
        new_node_metadata = node_metadata
    serialized_node_metadata = serialize_node_metadata(new_node_metadata)
    return new_graph_elements, serialized_node_metadata, metadata_str


def update_graph_with_search(search_text: str) -> Tuple[List[dict], Dict[str, RenkuObject]]:
    """
    Given a search term, search the Renku Knowledge Graph and construct graph elements to display the queried datasets.
    Args:
        search_text:

    Returns:
        Graph elements defining the nodes and edges of the returned matching datasets.
        Node metadata dictionary containing all nodes.
    """
    dataset_json = search_datasets_for_search_term(search_text)
    graph_elements, node_metadata = generate_graph_from_search_term(dataset_json, search_text)
    return graph_elements, node_metadata


def add_related_nodes_to_graph(node_id, graph_elements: List[dict], node_metadata: Dict
                               ) -> Tuple[List[dict], Dict[str, RenkuObject]]:
    """
        Given a node already in the graph, search the Renku Knowledge Graph for related datasets/projects and append
         graph elements to the currently display the queried datasets.
        Args:
            node_id: id of the node for which to add related datasets/projects
            graph_elements: graph elements currently depicted in the graph
            node_metadata: node metadata currently stored in the dashboard

        Returns:
            Graph elements defining the nodes and edges of the current graph plus the new related nodes
            Node metadata dictionary for all current nodes plus the new related nodes.
        """
    node_object = node_metadata[node_id]
    new_graph_elements, new_node_metadata = node_object.generate_related_network()
    graph_elements.extend(new_graph_elements)
    node_metadata.update(new_node_metadata)
    return graph_elements, node_metadata


def generate_graph_from_search_term(json: List[dict], keyword: str) -> Tuple[List[dict], Dict[str, RenkuObject]]:
    """
    Given the json describing datasets matching a search query, construct the graph elements and node_metadata to
     display the datasets
    Args:
        json: json result describing the datasets that matched the search term.
        keyword: The term that was searched. This is used to create a search term node and create graph edges between
         each dataset and the search term.

    Returns:
        * cytoscape graph elements definition
        * Dictionary of RenkuDataSets objects, one for each node containing all its metadata.
    """
    graph_elements = []
    node_metadata = {}

    graph_elements.append({'data': {'id': keyword, 'label': keyword}, 'classes': 'search_term'})
    for dataset in json:
        renku_dataset = RenkuDataSet.from_id(dataset['identifier'])
        graph_elements.append(renku_dataset.to_cytoscape_node())
        graph_elements.append({'data': {'source': renku_dataset.doi, 'target': keyword}})
        node_metadata[renku_dataset.doi] = renku_dataset

    return graph_elements, node_metadata


def serialize_node_metadata(node_metadata: Dict[str, RenkuObject]) -> List[Dict]:
    """
    Convert node_metadata object into a form that is json serializable for storage in the Dash Store component.

    Args:
        node_metadata: Dictionary of RenkuDataSet and RenkuProject objects for each of the nodes currently depicted in
         the graph.

    Returns: List of dictionaries representing each of the node objects.

    """
    serialized_nodes = [node_data.to_dict() for node_id, node_data in node_metadata.items()]
    return serialized_nodes


def deserialize_node_metadata(serialized_node_metadata: List[Dict]) -> Dict[str, RenkuObject]:
    """
    Covert a serialized representation of all RenkuDataSet and RenkuProject objects into live objects.

    Args:
        serialized_node_metadata: List of dictionaries representing each of the node objects.

    Returns: Dictionary of RenkuDataSet and RenkuProject objects, one for each node containing all its metadata.
    """
    deserialized_nodes = {}
    constructors = {
        'dataset': RenkuDataSet,
        'project': RenkuProject,
    }
    for item in serialized_node_metadata:
        constructor = constructors[item['type']]
        renku_object = constructor(item['data'])
        deserialized_nodes[renku_object.node_id] = renku_object
    return deserialized_nodes


def initialize_graph() -> Tuple[List[dict], dict]:
    """Create a simple initial graph"""
    graph_elements = [
        {'data': {'id': 'start', 'label': 'Search above to get started!'}},
    ]
    node_metadata = {}
    return graph_elements, node_metadata


starting_graph_elements, starting_node_metadata = initialize_graph()
app.layout = get_app_layout(starting_graph_elements, starting_node_metadata)


def main():
    app.run_server(debug=True)


if __name__ == '__main__':
    main()
